use tokio::sync::mpsc;

mod packet; use packet::{Packet, QueryPacket}; pub use packet::{Query, Response, Mode};
mod signal; pub use signal::Signal;
mod manager; use manager::{Manager, Client, manager_thread};

const BUFFER_SIZE :usize = 32;

pub struct Bus<T> where T: 'static + Clone + Send {
    sender:Option<mpsc::Sender<Packet<T>>>,
    receiver:Option<mpsc::Receiver<Packet<T>>>,
    query:mpsc::Sender<QueryPacket>,
    response:mpsc::Receiver<Response<T>>,
    id:u32,
}
impl<T: 'static + Clone + Send> Bus<T> {
    pub fn new() -> Self
    {
        let mut manager = Manager::<T>::new();
        let cq_sender = manager.query_sender.clone();
        let (cr_sender, cr_receiver) = mpsc::channel::<Response<T>>(BUFFER_SIZE);
        let id = manager.senders.add(Client {
            packet:None,
            query:cr_sender,
            signal:None,
        });
        
        tokio::spawn(manager_thread(manager));

        Self {
            sender:None,
            receiver:None,
            query:cq_sender,
            response:cr_receiver,
            id:id as u32,
        }
    }

    pub fn new_as(mode:Mode) -> Self
    {
        let transmitter :bool;
        let receiver :bool;
        match mode {
            Mode::Transmitter => {
                transmitter = true;
                receiver = false;
            }
            Mode::Transceiver => {
                transmitter = true;
                receiver = true;
            }
        }
        
        let mut manager = Manager::<T>::new();
        let sender = manager.sender.clone();
        let (c_sender, c_receiver) = mpsc::channel::<Packet<T>>(BUFFER_SIZE);
        let cq_sender = manager.query_sender.clone();
        let (cr_sender, cr_receiver) = mpsc::channel::<Response<T>>(BUFFER_SIZE);
        let id = manager.senders.add(Client {
            packet:if transmitter { Some(c_sender) } else { None },
            query:cr_sender,
            signal:None,
        });
        
        tokio::spawn(manager_thread(manager));

        Self {
            sender:if transmitter { Some(sender) } else { None },
            receiver:if receiver { Some(c_receiver) } else { None },
            query:cq_sender,
            response:cr_receiver,
            id:id as u32,
        }
    }

    pub async fn connect(&mut self) -> Result<Bus<T>,()>
    {
        self.connect_as(Mode::Transceiver).await
    }

    pub async fn connect_as(&mut self, mode:Mode) -> Result<Bus<T>,()>
    {
        if self.query.send(QueryPacket::new(self.id, Query::Connect(mode))).await.is_ok() {
            let recv = self.response.recv().await;
            match recv {
                Some(response) => {
                    match response {
                        Response::<T>::Connect(bus) => { return Ok(bus); }
                        _ => { }
                    }
                }
                _ => { }
            }
        }
        Err(())
    }

    pub fn disconnect(&self)
    // Remove the socket from the bus.
    {

    }

    pub async fn signal(&mut self, signal:Option<&Signal>) -> std::result::Result<(),()>
    {
        let sender :Option<mpsc::Sender<()>> = match signal {
            Some(signal) => Some(signal.register()),
            None => None,
        };
        if self.query.send(QueryPacket::new(self.id, Query::Signal(sender))).await.is_ok() {
            let recv = self.response.recv().await;
            match recv {
                Some(response) => {
                    match response {
                        Response::<T>::Ok => { return Ok(()); }
                        _ => { }
                    }
                }
                _ => { }
            }
        }
        Err(())
    }

    pub async fn send_from(&self, from:u32, to:u32, data:T) -> std::result::Result<(),()>
    // Send a packet to the bus.
    {
        match &self.sender {
            Some(transmitter) => {
                if transmitter.send(Packet::<T>::new(from, to, data)).await.is_ok() {
                    if self.query.send(QueryPacket::new(from, Query::Notify)).await.is_ok() {
                        return Ok(());
                    }
                }
            }
            None => { }
        }
        Err(())
    }

    pub async fn send(&self, to:u32, data:T) -> std::result::Result<(),()>
    // Send a packet to the bus.
    {
        return self.send_from(self.id, to, data).await;
    }

    pub fn receive(&mut self) -> Option<Packet<T>>
    // Return the next packet in the queue.
    {
        return match &mut self.receiver {
            Some(receiver) => {
                let result = receiver.try_recv();
                match result {
                    Ok(result) => Some(result),
                    Err(_) => None,
                }
            }
            None => None,
        };
    }

    pub async fn receive_wait(&mut self) -> Option<Packet<T>>
    // Return the next packet in the queue.
    {
        return match &mut self.receiver {
            Some(receiver) => {
                let result = receiver.recv().await;
                match result {
                    Some(result) => Some(result),
                    _ => None,
                }
            }
            None => None,
        };
    }

    pub fn receive_all(&mut self) -> Vec<Packet<T>>
    // Return all packets in the queue.
    {
        let mut results :Vec<Packet<T>> = vec![];
        match &mut self.receiver {
            Some(receiver) => {
                loop {
                    let result = receiver.try_recv();
                    match result {
                        Ok(packet) => {
                            results.push(packet);
                        }
                        Err(_) => { break; }
                    }
                }
            }
            None => { },
        }
        return results;
    }

    pub async fn set_mailbox(&mut self, id:u32) -> std::result::Result<(),()>
    // Map a mailbox number to this endpoint.
    {
        if self.query.send(QueryPacket::new(self.id, Query::SetPort(id))).await.is_ok() {
            let recv = self.response.recv().await;
            match recv {
                Some(response) => {
                    match response {
                        Response::<T>::Ok => { return Ok(()); }
                        _ => { }
                    }
                }
                _ => { }
            }
        }
        Err(())
    }

    pub async fn mailbox(&mut self, id:u32) -> Option<u32>
    // Return the endpoint address mapped to the specified mailbox number.
    {
        if self.query.send(QueryPacket::new(self.id, Query::GetPort(id))).await.is_ok() {
            let recv = self.response.recv().await;
            match recv {
                Some(response) => {
                    match response {
                        Response::<T>::GetPort(id) => { return id; }
                        _ => { }
                    }
                }
                _ => { }
            }
        }
        None
    }

    pub fn id(&self) -> u32
    {
        self.id
    }
}
